/*
 * Copyright (C) 2021, Xilinx Inc - All rights reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may
 * not use this file except in compliance with the License. A copy of the
 * License is located at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

#include "InternalSideData.h"
#include <cstring>

namespace xma {

#define XMA_SIDEDATA "SideData"

//////////////////////////////////////////////////////////////////////////////
InternalSideData::InternalSideData(IntSessionFactoryPtr sf, LogPtr log, enum XmaFrameSideDataType type, XmaBufferType buffer_type, uint64_t size) : m_sf(sf), m_log(log) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::InternalSideData(type=%d, buffer_type=%s, size=%lu)", type,
      ((buffer_type == XMA_HOST_BUFFER_TYPE)               ? "host" :
              (buffer_type == XMA_DEVICE_ONLY_BUFFER_TYPE) ? "device" :
              (buffer_type == XMA_DEVICE_BUFFER_TYPE)      ? "both" :
              (buffer_type == NO_BUFFER)                   ? "none" :
                                                             "unknown"),
      size);

  if ((buffer_type != XMA_HOST_BUFFER_TYPE) && (buffer_type != XMA_DEVICE_ONLY_BUFFER_TYPE) && (buffer_type != XMA_DEVICE_BUFFER_TYPE)) {
    THROW_INVALID_ARGUMENT_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::InternalSideData() Invalid XmaBufferType!");
  }
  m_sd = std::make_unique<XmaFrameSideData>(); // this will throw on OOM
  memset(m_sd.get(), 0, sizeof(XmaFrameSideData));

  m_sd->type        = type;
  m_sd->buffer_type = buffer_type;
  m_sd->on_host     = (buffer_type == XMA_HOST_BUFFER_TYPE) || (buffer_type == XMA_DEVICE_BUFFER_TYPE);
  buffer_t buffer   = m_sf->GetBuffer(size, buffer_type);
  m_sd->host        = static_cast<uint8_t*>(buffer->m_handle->virtual_address_rc);
  m_sd->device      = static_cast<uint64_t>(buffer->m_handle->bus_address_ep);
  m_sd->size        = size;

  auto info = std::make_unique<PrivateSideDataInfo>(m_sf, log, std::move(buffer));
  info->SetSelfPtr(this);
  m_sd->private_do_not_touch = static_cast<void*>(info.release());

  if (m_sd->on_host) {
    memset(m_sd->host, 0x00, size);
    if (buffer_type == XMA_DEVICE_BUFFER_TYPE) {
      Write();
    }
  }

#ifdef XMA_CHECK_MEM
  std::unique_lock<std::mutex> lock(m_sf->m_mutex);
  m_sf->m_side_data.push_back(m_sd.get());
  lock.unlock();
#endif

  IncRef();
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::InternalSideData() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
InternalSideData::~InternalSideData() {
  if (!m_sd) {
    return;
  }

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::~InternalSideData(type=%d, buffer_type=%s, size=%lu)", m_sd->type,
      ((m_sd->buffer_type == XMA_HOST_BUFFER_TYPE)               ? "host" :
              (m_sd->buffer_type == XMA_DEVICE_ONLY_BUFFER_TYPE) ? "device" :
              (m_sd->buffer_type == XMA_DEVICE_BUFFER_TYPE)      ? "both" :
              (m_sd->buffer_type == NO_BUFFER)                   ? "none" :
                                                                   "unknown"),
      m_sd->size);

  DecRef();

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::~InternalSideData() - Returning");
}

///////////////////////////////////////////////////////////////////////////////
XmaFrameSideData* InternalSideData::GetSideData() {
  LOCK_SIDE_DATA(m_log, "InternalSideData::GetSideData", XMA_SIDEDATA, m_sd);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::GetSideData(type=%d, buffer_type=%s, size=%lu)", m_sd->type,
      ((m_sd->buffer_type == XMA_HOST_BUFFER_TYPE)               ? "host" :
              (m_sd->buffer_type == XMA_DEVICE_ONLY_BUFFER_TYPE) ? "device" :
              (m_sd->buffer_type == XMA_DEVICE_BUFFER_TYPE)      ? "both" :
              (m_sd->buffer_type == NO_BUFFER)                   ? "none" :
                                                                   "unknown"),
      m_sd->size);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::GetSideData() - Returning");

  return m_sd.get();
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalSideData::IncRef() {
  LOCK_SIDE_DATA(m_log, "InternalSideData::IncRef", XMA_SIDEDATA, m_sd);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::IncRef(type=%d, buffer_type=%s, size=%lu)", m_sd->type,
      ((m_sd->buffer_type == XMA_HOST_BUFFER_TYPE)               ? "host" :
              (m_sd->buffer_type == XMA_DEVICE_ONLY_BUFFER_TYPE) ? "device" :
              (m_sd->buffer_type == XMA_DEVICE_BUFFER_TYPE)      ? "both" :
              (m_sd->buffer_type == NO_BUFFER)                   ? "none" :
                                                                   "unknown"),
      m_sd->size);

  ++m_sd->refcount;

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::IncRef() - Returning");
  return m_sd->refcount;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalSideData::DecRef() {
  LOCK_SIDE_DATA(m_log, "InternalSideData::DecRef", XMA_SIDEDATA, m_sd);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::DecRef(type=%d, buffer_type=%s, size=%lu)", m_sd->type,
      ((m_sd->buffer_type == XMA_HOST_BUFFER_TYPE)               ? "host" :
              (m_sd->buffer_type == XMA_DEVICE_ONLY_BUFFER_TYPE) ? "device" :
              (m_sd->buffer_type == XMA_DEVICE_BUFFER_TYPE)      ? "both" :
              (m_sd->buffer_type == NO_BUFFER)                   ? "none" :
                                                                   "unknown"),
      m_sd->size);

  --m_sd->refcount;

  if (m_sd->refcount > 0) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::DecRef() - Returning");
    return m_sd->refcount;
  }

  switch (m_sd->type) {
  case XMA_FRAME_SIDE_DATA_HDR10_PARAMS:
    if (m_sd->buffer_type == XMA_HOST_BUFFER_TYPE) {
      XmaHdrParams* hdr_params = reinterpret_cast<XmaHdrParams*>(m_sd->host);
      for (uint8_t idx = 0; idx < hdr_params->t35.payload_count; ++idx) {
        delete[] hdr_params->t35.payload_byte_data[idx];
      }
    }
    break;
  default:
    break;
  }

  m_sd->host        = nullptr;
  m_sd->device      = 0;
  m_sd->buffer_type = NO_BUFFER;

#ifdef XMA_CHECK_MEM
  std::unique_lock<std::mutex> lock(m_sf->m_mutex);
  m_sf->m_side_data.remove(m_sd.get());
  lock.unlock();
#endif

  if (m_sd->private_do_not_touch) {
    PrivateSideDataInfo* info = static_cast<PrivateSideDataInfo*>(m_sd->private_do_not_touch);
    if (info->IsValid()) {
      guard.unlock();
      delete info;
      m_sd->private_do_not_touch = nullptr;
    }
  }

  m_sd = nullptr;

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::DecRef() - Returning");

  return 0;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalSideData::GetRefCount() const {
  LOCK_SIDE_DATA(m_log, "InternalSideData::GetRefCount", XMA_SIDEDATA, m_sd);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::GetRefCount(type=%d, buffer_type=%s, size=%lu)", m_sd->type,
      ((m_sd->buffer_type == XMA_HOST_BUFFER_TYPE)               ? "host" :
              (m_sd->buffer_type == XMA_DEVICE_ONLY_BUFFER_TYPE) ? "device" :
              (m_sd->buffer_type == XMA_DEVICE_BUFFER_TYPE)      ? "both" :
              (m_sd->buffer_type == NO_BUFFER)                   ? "none" :
                                                                   "unknown"),
      m_sd->size);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::GetRefCount() - Returning");
  return m_sd->refcount;
}

///////////////////////////////////////////////////////////////////////////////
XmaFrameSideDataType InternalSideData::GetType() const {
  LOCK_SIDE_DATA(m_log, "InternalSideData::GetType", XMA_SIDEDATA, m_sd);

  return m_sd->type;
}

///////////////////////////////////////////////////////////////////////////////
XmaBufferType InternalSideData::GetBufferType() const {
  LOCK_SIDE_DATA(m_log, "InternalSideData::GetBufferType", XMA_SIDEDATA, m_sd);

  return m_sd->buffer_type;
}

///////////////////////////////////////////////////////////////////////////////
bool InternalSideData::IsOnHost() const {
  LOCK_SIDE_DATA(m_log, "InternalSideData::IsOnHost", XMA_SIDEDATA, m_sd);

  return m_sd->on_host;
}

///////////////////////////////////////////////////////////////////////////////
uint8_t* InternalSideData::GetHostPtr() const {
  LOCK_SIDE_DATA(m_log, "InternalSideData::GetHostPtr", XMA_SIDEDATA, m_sd);

  if ((m_sd->buffer_type != XMA_HOST_BUFFER_TYPE) && (m_sd->buffer_type != XMA_DEVICE_BUFFER_TYPE)) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::GetHostPtr() XmaFrameSideData does not contain a host buffer!");
  }

  return m_sd->host;
}

///////////////////////////////////////////////////////////////////////////////
uint64_t InternalSideData::GetDevicePtr() const {
  LOCK_SIDE_DATA(m_log, "InternalSideData::GetDevicePtr", XMA_SIDEDATA, m_sd);

  if ((m_sd->buffer_type != XMA_DEVICE_ONLY_BUFFER_TYPE) && (m_sd->buffer_type != XMA_DEVICE_BUFFER_TYPE)) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::GetDevicePtr() XmaFrameSideData does not contain a device buffer!");
  }

  return m_sd->device;
}

///////////////////////////////////////////////////////////////////////////////
size_t InternalSideData::GetSize() const {
  LOCK_SIDE_DATA(m_log, "InternalSideData::GetSize", XMA_SIDEDATA, m_sd);

  return m_sd->size;
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalSideData::Read() {
  LOCK_SIDE_DATA(m_log, "InternalSideData::Read", XMA_SIDEDATA, m_sd);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::Read(type=%d)", m_sd->type);

  if (m_sd->buffer_type != XMA_DEVICE_BUFFER_TYPE) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::Read() XmaFrameSideData does not contain both host and device buffers!");
  }

  if (m_sd->on_host) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::Read() - Returning");
    return XMA_SUCCESS;
  }

  if (m_sd->private_do_not_touch) {
    PrivateSideDataInfo* info = static_cast<PrivateSideDataInfo*>(m_sd->private_do_not_touch);
    if (info->IsValid()) {
      VpiLinearMem* mem = info->GetMem();
      if (vpi_memory_mem_sync(m_sf->GetMemHandle(), mem, mem, 0, m_sd->size, VPI_DATA_DWLOAD) < 0) {
        THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::Read() Read failed!");
      } else {
        m_sd->on_host = true;
        m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::Read() - Returning");
        return XMA_SUCCESS;
      }
    }
  }
  THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::Read() XmaFrameSideData was not created by xma_side_data_alloc()!");
}

///////////////////////////////////////////////////////////////////////////////
int32_t InternalSideData::Write() {
  LOCK_SIDE_DATA(m_log, "InternalSideData::Write", XMA_SIDEDATA, m_sd);

  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::Write(type=%d)", m_sd->type);

  if (m_sd->buffer_type != XMA_DEVICE_BUFFER_TYPE) {
    THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::Write() XmaFrameSideData does not contain both host and device buffers!");
  }

  if (!m_sd->on_host) {
    m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::Write() - Returning");
    return XMA_SUCCESS;
  }

  if (m_sd->private_do_not_touch) {
    PrivateSideDataInfo* info = static_cast<PrivateSideDataInfo*>(m_sd->private_do_not_touch);
    if (info->IsValid()) {
      VpiLinearMem* mem = info->GetMem();
      if (vpi_memory_mem_sync(m_sf->GetMemHandle(), mem, mem, 0, m_sd->size, VPI_DATA_UPLOAD) < 0) {
        THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::Write() Write failed!");
      } else {
        m_sd->on_host = false;
        m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "InternalSideData::Write() - Returning");
        return XMA_SUCCESS;
      }
    }
  }
  THROW_RUNTIME_EXCEPTION(m_log, XMA_ERROR_LOG, XMA_SIDEDATA, "InternalSideData::Write() XmaFrameSideData was not created by xma_side_data_alloc()!");
}

///////////////////////////////////////////////////////////////////////////////
int32_t AddSideDataToVpiFrm(LogPtr log, FramePtr src, VpiFrm* dst) {
  if (!src) {
    THROW_INVALID_ARGUMENT_EXCEPTION(log, XMA_ERROR_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm() No available source frame!");
  }
  if (!dst) {
    THROW_INVALID_ARGUMENT_EXCEPTION(log, XMA_ERROR_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm() No available destination frame!");
  }

  log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm()");

  for (int32_t i = 0; i < XMA_FRAME_SIDE_DATA_MAX_COUNT; ++i) {
    SideDataPtr sd = src->GetSideData(static_cast<XmaFrameSideDataType>(i));
    if (sd) {
      switch (sd->GetType()) {
      case XMA_FRAME_SIDE_DATA_DYN_ENC_PARAMS:
        if (sd->GetBufferType() != XMA_HOST_BUFFER_TYPE) {
          THROW_INVALID_ARGUMENT_EXCEPTION(log, XMA_ERROR_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm() Dynamic encoder params sidedata must be on host!");
        }
        memcpy(&dst->dynamic_params, sd->GetHostPtr(), sizeof(XmaDynamicEncParams));
        break;
      case XMA_FRAME_SIDE_DATA_HDR10_PARAMS: {
        if (sd->GetBufferType() != XMA_HOST_BUFFER_TYPE) {
          THROW_INVALID_ARGUMENT_EXCEPTION(log, XMA_ERROR_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm() HDR sidedata must be on host!");
        }
        dst->hdr_param = new VpiHdrParameters;
        if (!dst->hdr_param) {
          THROW_BAD_ALLOC_EXCEPTION(log, XMA_ERROR_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm() Out of memory!");
        }
        if (sd->GetSize() != sizeof(XmaHdrParams)) {
          THROW_INVALID_ARGUMENT_EXCEPTION(
              log, XMA_ERROR_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm() HDR sidedata is incorrect size, must by %lu bytes!", sizeof(XmaHdrParams));
        }
        if (sizeof(VpiHdrParameters) != sizeof(XmaHdrParams)) {
          THROW_RUNTIME_EXCEPTION(log, XMA_ERROR_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm() HDR sidedata is incorrect size, must be same size as VSI HDR sidedata!");
        }
        XmaHdrParams* hdr_params = reinterpret_cast<XmaHdrParams*>(sd->GetHostPtr());
        memcpy(dst->hdr_param, hdr_params, sizeof(VpiHdrParameters));
        for (uint8_t idx = 0; idx < hdr_params->t35.payload_count; ++idx) {
          dst->hdr_param->t35_param.payload_byte_data[idx] = new uint8_t[hdr_params->t35.payload_byte_length[idx]];
          if (!dst->hdr_param->t35_param.payload_byte_data[idx]) {
            THROW_BAD_ALLOC_EXCEPTION(log, XMA_ERROR_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm() Out of memory!");
          }
          memcpy(dst->hdr_param->t35_param.payload_byte_data, hdr_params->t35.payload_byte_data[idx], hdr_params->t35.payload_byte_length[idx]);
        }
        break;
      }
      case XMA_FRAME_SIDE_DATA_VUI_PARAMS:
        if (sd->GetBufferType() != XMA_HOST_BUFFER_TYPE) {
          THROW_INVALID_ARGUMENT_EXCEPTION(log, XMA_ERROR_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm() VUI sidedata must be on host!");
        }
        memcpy(&dst->vui_param, sd->GetHostPtr(), sizeof(XmaVuiParams));
        break;
      default:
        log->LogMsg(XMA_WARNING_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm() side data %d getting lost!", i);
        break;
      }
    }
  }

  log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "AddSideDataToVpiFrm() - Returning");

  return XMA_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////
PrivateSideDataInfo::PrivateSideDataInfo(IntSessionFactoryPtr sf, LogPtr log, buffer_t buffer) : m_buffer(std::move(buffer)), m_sf(sf), m_log(log) {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "PrivateSideDataInfo::PrivateSideDataInfo()");
}

PrivateSideDataInfo::~PrivateSideDataInfo() {
  m_log->LogMsg(XMA_DEBUG_LOG, XMA_SIDEDATA, "PrivateSideDataInfo::~PrivateSideDataInfo()");
}

bool PrivateSideDataInfo::IsValid() const {
  return (m_dummy == SIDE_DATA_DUMMY_VALUE);
}

std::recursive_mutex& PrivateSideDataInfo::GetMutex() {
  return m_mutex;
}

VpiLinearMem* PrivateSideDataInfo::GetMem() const {
  return m_buffer->m_handle;
}

void PrivateSideDataInfo::SetSelf(IntSideDataPtr self) {
  m_self = self;
}

IntSideDataPtr PrivateSideDataInfo::GetSelf() const {
  return m_self;
}

void PrivateSideDataInfo::SetSelfPtr(InternalSideData* self_ptr) {
  m_self_ptr = self_ptr;
}

InternalSideData* PrivateSideDataInfo::GetSelfPtr() const {
  return m_self_ptr;
}

IntSessionFactoryPtr PrivateSideDataInfo::GetSessionFactory() const {
  return m_sf;
}

LogPtr PrivateSideDataInfo::GetLog() const {
  return m_log;
}

}; // Namespace xma
