/*
 * Copyright (C) 2021, Xilinx Inc - All rights reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You may
 * not use this file except in compliance with the License. A copy of the
 * License is located at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations
 * under the License.
 */

#pragma once

#include <deque>
#include <mutex>
#include <functional>
#include "xmacpp/EncoderSession.h"
#include "SessionBase.h"
#include "InternalDataBuffer.h"

namespace xma {

#define XAV1_EXTRA_FRAMES 60
/* The delay will be relative to the frame rate, so a framerate of 60 fps and
 * retry delay of 10 will lead to a delay of 1/(60*10) seconds */
#define XMA_ENCODER_RETRY_DELAY 10
/* The retry count is relative to the retry delay (which is relative to the
 * framerate), so if the framerate is 60 fps and the retry count is
 * (XMA_ENCODER_RETRY_DELAY*100), the maximum delay will be 100*1/60
 * seconds */
#define XMA_ENCODER_MAX_RETRY_COUNT (XMA_ENCODER_RETRY_DELAY * 100)

//////////////////////////////////////////////////////////////////////////////
class InternalEncoderSession final : virtual public EncoderSession, virtual public SessionBase {
public:
  InternalEncoderSession(IntSessionFactoryPtr sf, XmaEncoderProperties* props);
  virtual ~InternalEncoderSession();

  virtual int32_t SetLog(LogPtr log) override;

  virtual int32_t Enqueue(queue_t& data) override;
  virtual int32_t Dequeue(queue_t& data) override;

private:
  void             InitParams(XmaEncoderProperties* props);
  void             StartEncoder(FramePtr frame);
  int32_t          GetPacket(IntDataBufferPtr& buf);
  IntDataBufferPtr PullPacket(int32_t* ret);

protected:
  using lock_t           = std::lock_guard<std::mutex>;
  using vpi_mem_handle_t = std::unique_ptr<void, std::function<void(VpiMemHandle mem_ctx)>>;
  using enc_handle_t     = std::unique_ptr<void, std::function<void(void*)>>;
  using pool_handle_t    = std::unique_ptr<void, std::function<void(void*)>>;
  std::mutex                   m_mutex;
  vpi_mem_handle_t             m_mem_ctx;
  enc_handle_t                 m_handle;
  VpiEncInitParams             m_params;
  char                         m_param_str[4096];
  uint8_t*                     m_extra_data      = nullptr;
  int32_t*                     m_extra_data_size = nullptr;
  std::deque<IntDataBufferPtr> m_queue;
  uint32_t                     m_max_frames_processing = 0;
  uint32_t                     m_frames_processing     = 0;
  int32_t                      m_wait                  = 0;
  bool                         m_end_found             = false;
  int64_t                      m_frame_sent            = 0;
  int64_t                      m_frame_recv            = 0;
  bool                         m_latency_logging       = false;
  struct timespec              m_latency;
  int64_t                      m_time_taken;
};

using IntEncoderSessionPtr = std::shared_ptr<InternalEncoderSession>;

}; // Namespace xma
